#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
#  rule_engine/engine.py
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are
#  met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions and the following disclaimer
#    in the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of the project nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
#  A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
#  OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
#  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
#  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import collections
import collections.abc
import datetime
import functools
import inspect
import math

from . import ast
from . import errors
from . import parser

import dateutil.tz

def _now(builtins):
	return datetime.datetime.now(tz=builtins.timezone)

def _today(builtins):
	return _now(builtins).replace(hour=0, minute=0, second=0, microsecond=0)

def resolve_attribute(thing, name):
	"""
	A replacement resolver function for looking up symbols as members of
	*thing*. This is effectively the same as ``thing.name``. The *thing* object
	can be a :py:class:`~collections.namedtuple`, a custom Python class or any
	other object. Each of the members of *thing* must be of a compatible data
	type.

	.. warning::
		This effectively exposes all members of *thing*. If any members are
		sensitive, then a custom resolver should be used that checks *name*
		against a whitelist of attributes that are allowed to be accessed.

	:param thing: The object on which the *name* attribute will be accessed.
	:param str name: The symbol name that is being resolved.
	:return: The value for the corresponding attribute *name*.
	"""
	if not hasattr(thing, name):
		raise errors.SymbolResolutionError(name, thing=thing)
	return getattr(thing, name)

def resolve_item(thing, name):
	"""
	A resolver function for looking up symbols as items from an object (*thing*)
	which supports the :py:class:`~collections.abc.Mapping` interface, such as a
	dictionary. This is effectively the same as ``thing['name']``. Each of the
	values in *thing* must be of a compatible data type.

	:param thing: The object from which the *name* item will be accessed.
	:param str name: The symbol name that is being resolved.
	:return: The value for the corresponding attribute *name*.
	"""
	if not isinstance(thing, collections.abc.Iterable):
		raise errors.SymbolResolutionError(name, thing=thing)
	if name not in thing:
		raise errors.SymbolResolutionError(name, thing=thing)
	return thing[name]

def _type_resolver(type_map, name):
	if name not in type_map:
		raise errors.SymbolResolutionError(name)
	return type_map[name]

def type_resolver_from_dict(dictionary):
	"""
	Return a function suitable for use as the *type_resolver* for a
	:py:class:`.Context` instance from a dictionary. If any of the values within
	the dictionary are not of a compatible data type, a :py:exc:`TypeError` will
	be raised. Additionally, the resulting function will raise a
	:py:exc:`~rule_engine.errors.SymbolResolutionError` if the symbol name does
	not exist within the dictionary.

	:param dict dictionary: A dictionary (or any other object which supports the
		:py:class:`~collections.abc.Mapping` interface) from which to create the
		callback function.
	:return: The callback function.
	:rtype: function
	"""
	type_map = {key: value if isinstance(value, ast.DataType) else ast.DataType.from_value(value) for key, value in dictionary.items()}
	return functools.partial(_type_resolver, type_map)

_AttributeResolverFunction = collections.namedtuple('_AttributeResolverFunction', ('function', 'result_type'))
class _AttributeResolver(object):
	class attribute(object):
		__slots__ = ('types', 'name', 'result_type')
		type_map = collections.defaultdict(dict)
		def __init__(self, name, *data_types, result_type=ast.DataType.UNDEFINED):
			self.types = data_types
			self.name = name
			self.result_type = result_type

		def __call__(self, function):
			for type_ in self.types:
				self.type_map[type_][self.name] = _AttributeResolverFunction(function, self.result_type)
			return function

	def __call__(self, thing, object_, name):
		try:
			object_type = ast.DataType.from_value(object_)
		except TypeError:
			# if the object can't be mapped to a supported type, raise a resolution error
			raise errors.AttributeResolutionError(name, object_, thing=thing) from None
		resolver = self._get_resolver(object_type, name, thing=thing)
		value = resolver.function(self, object_)
		value = ast.coerce_value(value)
		if resolver.result_type is ast.DataType.UNDEFINED:
			return value
		value_type = ast.DataType.from_value(value)
		if resolver.result_type is value_type:
			return value
		raise errors.AttributeTypeError(name, object_, is_value=value, is_type=value_type, expected_type=resolver.result_type)

	def _get_resolver(self, object_type, name, thing=errors.UNDEFINED):
		attribute_resolvers = self.attribute.type_map.get(object_type)
		if attribute_resolvers is None:
			raise errors.AttributeResolutionError(name, object_type, thing=thing)
		resolver = attribute_resolvers.get(name)
		if resolver is None:
			raise errors.AttributeResolutionError(name, object_type, thing=thing)
		return resolver

	def resolve_type(self, object_type, name):
		return self._get_resolver(object_type, name).result_type

	@attribute('date', ast.DataType.DATETIME, result_type=ast.DataType.DATETIME)
	def datetime_date(self, value):
		return value.replace(hour=0, minute=0, second=0, microsecond=0)

	@attribute('day', ast.DataType.DATETIME, result_type=ast.DataType.FLOAT)
	def datetime_day(self, value):
		return value.day

	@attribute('hour', ast.DataType.DATETIME, result_type=ast.DataType.FLOAT)
	def datetime_hour(self, value):
		return value.hour

	@attribute('microsecond', ast.DataType.DATETIME, result_type=ast.DataType.FLOAT)
	def datetime_microsecond(self, value):
		return value.microsecond

	@attribute('millisecond', ast.DataType.DATETIME, result_type=ast.DataType.FLOAT)
	def datetime_millisecond(self, value):
		return value.microsecond / 1000

	@attribute('minute', ast.DataType.DATETIME, result_type=ast.DataType.FLOAT)
	def datetime_minute(self, value):
		return value.minute

	@attribute('month', ast.DataType.DATETIME, result_type=ast.DataType.FLOAT)
	def datetime_month(self, value):
		return value.month

	@attribute('second', ast.DataType.DATETIME, result_type=ast.DataType.FLOAT)
	def datetime_second(self, value):
		return value.second

	@attribute('weekday', ast.DataType.DATETIME, result_type=ast.DataType.STRING)
	def datetime_weekday(self, value):
		# use strftime %A so the value is localized
		return value.strftime('%A')

	@attribute('year', ast.DataType.DATETIME, result_type=ast.DataType.FLOAT)
	def datetime_year(self, value):
		return value.year

	@attribute('zone_name', ast.DataType.DATETIME, result_type=ast.DataType.STRING)
	def datetime_zone_name(self, value):
		return value.tzname()

	@attribute('as_lower', ast.DataType.STRING, result_type=ast.DataType.STRING)
	def string_as_lower(self, value):
		return value.lower()

	@attribute('as_upper', ast.DataType.STRING, result_type=ast.DataType.STRING)
	def string_as_upper(self, value):
		return value.upper()

	@attribute('is_empty', ast.DataType.ARRAY, ast.DataType.STRING, result_type=ast.DataType.BOOLEAN)
	def value_is_empty(self, value):
		return len(value) == 0

	@attribute('length', ast.DataType.ARRAY, ast.DataType.STRING, result_type=ast.DataType.FLOAT)
	def value_length(self, value):
		return len(value)

class Builtins(collections.abc.Mapping):
	"""
	A class to define and provide variables to within the builtin context of
	rules. These can be accessed by specifying a symbol name with the ``$``
	prefix."""
	def __init__(self, values, namespace=None, timezone=None):
		"""
		:param dict values: A mapping of string keys to be used as symbol names
			with values of either Python literals or a function which will be
			called when the symbol is accessed. When using a function, it will
			be passed a single argument, which is the instance of
			:py:class:`Builtins`.
		:param str namespace: The namespace of the variables to resolve.
		:param timezone: A timezone to use when resolving timestamps.
		:type timezone: :py:class:`~datetime.tzinfo`
		"""
		self.__values = values
		self.namespace = namespace
		self.timezone = timezone or dateutil.tz.tzlocal()

	def __repr__(self):
		return "<{} namespace={!r} keys={!r} timezone={!r} >".format(self.__class__.__name__, self.namespace, tuple(self.keys()), self.timezone)

	def __getitem__(self, name):
		value = self.__values[name]
		if isinstance(value, collections.abc.Mapping):
			if self.namespace is None:
				namespace = name
			else:
				namespace = self.namespace + '.' + name
			return self.__class__(value, namespace=namespace, timezone=self.timezone)
		elif inspect.isfunction(value):
			value = value(self)
		return value

	def __iter__(self):
		return iter(self.__values)

	def __len__(self):
		return len(self.__values)

	@classmethod
	def from_defaults(cls, values=None, **kwargs):
		"""Initialize a :py:class:`Builtins` instance with a set of default values."""
		default_values = {
			'e': math.e,
			'pi': math.pi,
			'now': _now,
			'today': _today
		}
		if values is not None:
			default_values.update(values)
		return cls(default_values, **kwargs)

class Context(object):
	"""
	An object defining the context for a rule's evaluation. This can be used to
	change the behavior of certain aspects of the rule such as how symbols are
	resolved and what regex flags should be used.
	"""
	def __init__(self, regex_flags=0, resolver=None, type_resolver=None, default_timezone='local', default_value=errors.UNDEFINED):
		"""
		:param int regex_flags: The flags to provide to functions in the
			:py:mod:`re` module.
		:param resolver: An optional callback function to use in place of
			:py:meth:`.resolve`.
		:param type_resolver: An optional callback function to use in place of
			:py:meth:`.resolve_type`.
		:param default_timezone: The default timezone to apply to
			:py:class:`~datetime.datetime` instances which do not have one
			specified. This is necessary for comparison operations. The value
			should either be a :py:class:`~datetime.tzinfo` instance, or a
			string. If *default_timzezone* is a string it must be one of the
			specially supported (case-insensitive) values of "local" or "utc".
		:type default_timezone: str, :py:class:`~datetime.tzinfo`
		:param default_value: The default value to return when resolving either
			a missing symbol or attribute.

		.. versionchanged: v2.0.0
			Added the *default_value* parameter.
		"""
		self.regex_flags = regex_flags
		"""
		The flags to provide to the :py:func:`~re.match` and
		:py:func:`~re.search` functions when matching or searching for patterns.
		"""
		self.symbols = set()
		"""
		The symbols that are referred to by the rule. Some or all of these will
		need to be resolved at evaluation time. This attribute can be used after
		a rule is generated to ensure that all symbols are valid before it is
		evaluated.
		"""
		if isinstance(default_timezone, str):
			default_timezone = default_timezone.lower()
			if default_timezone == 'local':
				default_timezone = dateutil.tz.tzlocal()
			elif default_timezone == 'utc':
				default_timezone = dateutil.tz.tzutc()
			else:
				raise ValueError('unsupported timezone: ' + default_timezone)
		elif not isinstance(default_timezone, datetime.tzinfo):
			raise TypeError('invalid default_timezone type')
		self.default_timezone = default_timezone
		self.default_value = default_value
		self.builtins = Builtins.from_defaults(timezone=default_timezone)
		"""An instance of :py:class:`Builtins` to provided a default set of builtin symbol values."""
		self.__type_resolver = type_resolver or (lambda _: ast.DataType.UNDEFINED)
		self.__resolver = resolver or resolve_item

	def resolve(self, thing, name, scope=None):
		"""
		The method to use for resolving symbols names to values. This function
		must return a compatible value for the specified symbol name. When a
		*scope* is defined, this function handles the resolution itself, however
		when the *scope* is ``None`` the resolver specified in
		:py:meth:`~.Context.__init__` is used which defaults to
		:py:func:`resolve_item`.

		:param thing: The object from which the *name* item will be accessed.
		:param str name: The symbol name that is being resolved.
		:return: The value for the corresponding symbol *name*.
		"""
		if scope == 'built-in':
			thing = self.builtins
		if isinstance(thing, Builtins):
			return resolve_item(thing, name)
		if scope is None:
			try:
				return self.__resolver(thing, name)
			except errors.SymbolResolutionError:
				if self.default_value is not errors.UNDEFINED:
					return self.default_value
				raise
		raise errors.SymbolResolutionError(name, symbol_scope=scope, thing=thing)

	__resolve_attribute = _AttributeResolver()
	def resolve_attribute(self, thing, object_, name):
		"""
		The method to use for resolving attributes from values. This function
		must return a compatible value for the specified attribute name.

		:param thing: The object from which the *object_* was retrieved.
		:param object_: The object from which the *name* attribute will be accessed.
		:param str name: The attribute name that is being resolved.
		:return: The value for the corresponding attribute *name*.
		"""
		try:
			return self.__resolve_attribute(thing, object_, name)
		except errors.AttributeResolutionError:
			if self.default_value is not errors.UNDEFINED:
				return self.default_value
			raise
	resolve_attribute_type = __resolve_attribute.resolve_type

	def resolve_type(self, name):
		"""
		A method for providing type hints while the rule is being generated.
		This can be used to ensure that all symbol names are valid and that the
		types are appropriate for the operations being performed. It must then
		return one of the compatible data type constants if the symbol is valid
		or raise an exception. The default behavior is to return
		:py:data:`~rule_engine.ast.DataType.UNDEFINED` for all symbols.

		:param str name: The symbol name to provide a type hint for.
		:return: The type of the specified symbol
		"""
		return self.__type_resolver(name)

class Rule(object):
	"""
	A rule which parses a string with a logical expression and can then evaluate
	an arbitrary object for whether or not it matches based on the constraints
	of the expression.
	"""
	parser = parser.Parser()
	"""
	The :py:class:`~rule_engine.parser.Parser` instance that will be used for
	parsing the rule text into a compatible abstract syntax tree (AST) for
	evaluation.
	"""
	def __init__(self, text, context=None):
		"""
		:param str text: The text of the logical expression.
		:param context: The context to use for evaluating the expression on
			arbitrary objects. This can be used to change the default behavior.
			The default context is :py:class:`.Context` but any object providing
			the same interface (such as a subclass) can be used.
		:type context: :py:class:`.Context`
		"""
		context = context or Context()
		self.text = text
		self.context = context
		self.statement = self.parser.parse(text, context)

	def __repr__(self):
		return "<{0} text={1!r} >".format(self.__class__.__name__, self.text)

	def __str__(self):
		return self.text

	def filter(self, things):
		"""
		A convenience function for iterating over *things* and yielding each
		member that :py:meth:`.matches` return True for.

		:param things: The collection of objects to iterate over.
		"""
		yield from (thing for thing in things if self.matches(thing))

	@classmethod
	def is_valid(cls, text, context=None):
		"""
		Test whether or not the rule is syntactically correct. This verifies the
		grammar is well structured and that there are no type compatibility
		issues regarding literals or symbols with known types (see
		:py:meth:`~.Context.resolve_type` for specifying symbol type
		information).

		:param str text: The text of the logical expression.
		:param context: The context as would be passed to the
			:py:meth:`.__init__` method. This can be used for specifying symbol
			type information.
		:return: Whether or not the expression is well formed and appears valid.
		:rtype: bool
		"""
		try:
			cls.parser.parse(text, (context or Context()))
		except errors.EngineError:
			return False
		return True

	def evaluate(self, thing):
		"""
		Evaluate the rule against the specified *thing* and return the value.
		This can be used to, for example, apply the symbol resolver.

		:param thing: The object on which to apply the rule.
		:return: The value the rule evaluates to. Unlike the :py:meth:`.matches`
			method, this is not necessarily a boolean.
		"""
		return self.statement.evaluate(thing)

	def matches(self, thing):
		"""
		Evaluate the rule against the specified *thing*. This will either return
		whether *thing* matches, or an exception will be raised.

		:param thing: The object on which to apply the rule.
		:return: Whether or not the rule matches.
		:rtype: bool
		"""
		return bool(self.statement.evaluate(thing))

	def to_graphviz(self):
		"""
		Generate a diagram of the parsed rule's AST using GraphViz.

		:return: The rule diagram.
		:rtype: :py:class:`graphviz.Digraph`
		"""
		import graphviz
		digraph = graphviz.Digraph(comment=self.text)
		self.statement.to_graphviz(digraph)
		return digraph

class DebugRule(Rule):
	parser = None
	def __init__(self, *args, **kwargs):
		self.parser = parser.Parser(debug=True)
		super(DebugRule, self).__init__(*args, **kwargs)
